import itertools
import json


def pascal_case(s):
    # E.g. "snake_case" -> "SnakeCase"
    return s.replace("_", " ").title().replace(" ", "")


def sorted_groupby(iterable, key):
    return itertools.groupby(sorted(iterable, key=key), key=key)


def compact_dumps(data):
    return json.dumps(data, separators=(",", ":"))



import enum

from routes import const


class ResultType(enum.Enum):
    ROUTE = "route"
    CONNECTOR = "connector"


class Framework(enum.Enum):
    FLASK = "flask"
    DJANGO = "django"
    DJANGO_REST_FRAMEWORK = "django-rest-framework"
    SANIC = "sanic"
    LARAVEL = "laravel"
    SYMFONY = "symfony"
    CAKEPHP = "cakephp"
    RAILS = "rails"
    GRAPE = "grape"


class SemgrepResult:
    def __init__(self, result):
        self.result = result

    @property
    def check_id(self):
        return self.result["check_id"]

    @property
    def path(self):
        return self.result["path"]

    @property
    def start_line(self):
        return self.result["start"]["line"]

    @property
    def lines(self):
        return self.result["extra"]["lines"]

    @property
    def first_line(self):
        newline = self.lines.find("\n")
        return self.lines[:newline] if newline != -1 else self.lines

    @property
    def metadata(self):
        return self.result["extra"].get("metadata", {})

    @property
    def metavars(self):
        return self.result["extra"]["metavars"]

    def metavar_content(self, metavar):
        return self.metavars[metavar]["abstract_content"]

    @property
    def rd_metadata(self):
        return self.metadata.get("route-detect", {})

    @property
    def rd_type(self):
        return self.rd_metadata.get("type", ResultType.ROUTE.value)

    @property
    def rd_normalizer(self):
        return self.rd_metadata.get("normalizer")

    @property
    def rd_fill(self):
        return self.rd_metadata.get("fill", const.DEFAULT_FILL_COLOR)

    @property
    def rd_connect_on(self):
        return self.rd_metadata["connect-on"]



import argparse
import logging
import sys

from routes import __version__
from routes import commands
from routes import rules
from routes import templates

logger = logging.getLogger(__name__)


def parse_args(args=None):
    p = argparse.ArgumentParser(
        description="Find web application HTTP route authn and authz security bugs in your code.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    p.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Set logging level to debug",
    )
    p.add_argument(
        "-V",
        "--version",
        action="store_true",
        help="Print version information",
    )

    subparsers = p.add_subparsers(dest="command", help="Command help")

    which_parser = subparsers.add_parser(
        "which", help="Output path of supported Semgrep rule"
    )
    which_parser.add_argument(
        "rule", choices=list(rules.ALL_RULES.keys()), help="Semgrep rule"
    )

    viz_parser = subparsers.add_parser("viz", help="Route visualization tool")
    viz_parser.add_argument(
        "input",
        action="store",
        type=argparse.FileType("r"),
        help="Semgrep JSON output file",
    )
    viz_parser.add_argument(
        "-o",
        "--output",
        action="store",
        default=templates.ROUTES_HTML,
        type=argparse.FileType("w"),
        help="Routes HTML output file",
    )
    viz_parser.add_argument(
        "-t",
        "--template",
        action="store",
        default=str(templates.ROUTES_HTML_TEMPLATE),
        type=argparse.FileType("r"),
        help="Routes HTML template file",
    )
    viz_parser.add_argument(
        "--browser",
        action="store_true",
        help="Open HTML output file in browser",
    )

    return p.parse_args(args=args)


def main():
    args = parse_args()

    if args.version:
        print(__version__)
        return 0

    logging.basicConfig(
        format="%(asctime)s %(levelname)s %(name)s %(message)s",
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose else logging.INFO,
    )

    command_dispatch = {"which": commands.which.main, "viz": commands.viz.main}
    command = command_dispatch.get(args.command)
    if command is None:
        logger.error(f"Command unavailable: {args.command}")
        return 1

    logger.info("Starting command %s", args.command)
    result = command(args)
    logger.info("Finished command %s", args.command)

    return result


if __name__ == "__main__":
    sys.exit(main())



import collections
import json
import logging
import pathlib
import re
import webbrowser

from routes import const
from routes import types
from routes import util

logger = logging.getLogger(__name__)

RAIL_SYMBOL_RE = re.compile(r":(?P<name>[a-z_]+)")
RAIL_CONTROLLER_RE = re.compile(r"(?P<controller>[a-z_]+)#(?P<action>[a-z_]+)")


def rails_route_to_controller(result):
    content = result.metavar_content(result.rd_connect_on)

    # Symbol route, e.g. ":user"
    symbol_match = re.match(RAIL_SYMBOL_RE, content)
    if symbol_match is not None:
        return util.pascal_case(symbol_match.group("name")) + "Controller"

    # Controller name, e.g. "user#action"
    controller_match = re.search(RAIL_CONTROLLER_RE, content)
    if controller_match is not None:
        return util.pascal_case(controller_match.group("controller")) + "Controller"

    # Else try content itself even though it's unlikely to work
    return content


NORMALIZERS = {types.Framework.RAILS.value: rails_route_to_controller}


def get_connectors(connector_results):
    def connector_key(result):
        return result.metavar_content(result.rd_connect_on)

    results = {}
    for key, group in util.sorted_groupby(connector_results, key=connector_key):
        group_list = list(group)
        if len(group_list) > 1:
            logger.warning("Grouping on %s is ambiguous", key)
            continue
        results[key] = group_list[0]

    return results


def d3ify(parts, output, result, connectors):
    part = parts.pop(0)

    new_node = {"name": part}

    if parts:
        new_output = []
        new_node["children"] = new_output
        d3ify(parts, new_output, result, connectors)
    else:
        name = f"ln {result.start_line}: {result.first_line}"

        if result.rd_normalizer:
            normalizer = NORMALIZERS.get(result.rd_normalizer)
            normalized = normalizer(result)
            connector = connectors.get(normalized)
            fill = connector.rd_fill if connector else result.rd_fill
        else:
            fill = result.rd_fill

        check_node = {"name": name, "fill": fill}
        new_node.setdefault("children", []).append(check_node)

    output.append(new_node)


def merge_d3_results(d1s, d2s):
    for d2 in d2s:
        matching_d1 = next((d1 for d1 in d1s if d1["name"] == d2["name"]), None)
        if matching_d1 is None or "children" not in d2:
            d1s.append(d2)
        elif "children" in matching_d1 and "children" in d2:
            merge_d3_results(matching_d1["children"], d2["children"])


def main(args):
    logger.info("Reading input file %s", args.input.name)
    data = json.load(args.input)

    semgrep_results = [types.SemgrepResult(r) for r in data["results"]]
    counts = collections.Counter([r.check_id for r in semgrep_results])
    count_output = " ".join(f"{k}={v}" for k, v in counts.items())
    logger.info("Finding rule counts: %s", count_output)

    results_by_type = {
        key: list(group)
        for key, group in util.sorted_groupby(semgrep_results, key=lambda r: r.rd_type)
    }

    connector_results = results_by_type.get(types.ResultType.CONNECTOR.value, {})
    connectors = get_connectors(connector_results)

    root_paths = set()
    d3_results = []
    for result in results_by_type.get(types.ResultType.ROUTE.value, []):
        path = pathlib.PurePath(result.path)
        logger.debug("Processing %s %s %s", result.check_id, path, result.start_line)
        root, *_ = path.parts
        root_paths.add(root)
        output = []
        d3ify(list(path.parts), output, result, connectors)
        d3_results.append(output)

    all_same_root = len(root_paths) == 1
    if root_paths and not all_same_root:
        raise ValueError(
            f"Tree assumes a common root ({root_paths}), please only specify a single directory"
        )

    d3_tree = []
    for d3_result in d3_results:
        merge_d3_results(d3_tree, d3_result)

    # Since all paths share the same root we can start our tree there
    d3_tree = d3_tree[0] if d3_tree else {}

    if not d3_tree:
        logging.error(
            "No results found, please ensure your framework is supported or "
            "file an issue at https://github.com/mschwager/route-detect/issues"
        )

    template_name = pathlib.PurePath(args.template.name).name
    logger.info("Formatting template %s", template_name)
    output_buff = args.template.read()
    template_data = util.compact_dumps(d3_tree)
    output_buff = output_buff.replace(const.DEFAULT_TEMPLATE_KEY, template_data)

    logger.info("Writing output file %s", args.output.name)
    written = args.output.write(output_buff)
    success = len(output_buff) == written

    if args.browser:
        logger.info("Opening output file in browser")
        output_uri = pathlib.Path(args.output.name).resolve().as_uri()
        webbrowser.open(output_uri)

    return 0 if success else 1



import logging

from routes import rules

logger = logging.getLogger(__name__)


def main(args):
    logger.info("Printing rule path for %s", args.rule)
    print(rules.ALL_RULES[args.rule], end="")
    return 0



import pathlib


CURDIR = pathlib.Path(__file__).parent
TEST_ROUTE_DETECT = CURDIR / "test-route-detect.yml"
FLASK = CURDIR / "flask.yml"
DJANGO = CURDIR / "django.yml"
DJANGO_REST_FRAMEWORK = CURDIR / "django-rest-framework.yml"
SANIC = CURDIR / "sanic.yml"
LARAVEL = CURDIR / "laravel.yml"
SYMFONY = CURDIR / "symfony.yml"
CAKEPHP = CURDIR / "cakephp.yml"
RAILS = CURDIR / "rails.yml"
GRAPE = CURDIR / "grape.yml"
JAXRS = CURDIR / "jax-rs.yml"
SPRING = CURDIR / "spring.yml"
GORILLA = CURDIR / "gorilla.yml"
GIN = CURDIR / "gin.yml"
CHI = CURDIR / "chi.yml"
EXPRESS = CURDIR / "express.yml"
REACT = CURDIR / "react.yml"
ANGULAR = CURDIR / "angular.yml"

ALL_RULES = {
    str(rule.with_suffix("").name): str(rule.resolve())
    for rule in [
        TEST_ROUTE_DETECT,
        FLASK,
        DJANGO,
        DJANGO_REST_FRAMEWORK,
        SANIC,
        LARAVEL,
        SYMFONY,
        CAKEPHP,
        RAILS,
        GRAPE,
        JAXRS,
        SPRING,
        GORILLA,
        GIN,
        CHI,
        EXPRESS,
        REACT,
        ANGULAR,
    ]
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['routes', 'routes.commands', 'routes.rules', 'routes.templates']

package_data = \
{'': ['*']}

install_requires = \
['semgrep>=1.9.0,<2.0.0']

entry_points = \
{'console_scripts': ['routes = routes.main:main']}

setup_kwargs = {
    'name': 'route-detect',
    'version': '0.5.0',
    'description': 'Find web application HTTP route authn and authz security bugs in your code.',
    'long_description': '# route-detect\n\nWeb application HTTP route authentication (authn) and authorization (authz) bugs are some of the most common security issues found today. These industry standard resources highlight the severity of the issue:\n\n- 2021 OWASP Top 10 #1 - [Broken Access Control](https://owasp.org/Top10/A01_2021-Broken_Access_Control/)\n- 2021 OWASP Top 10 #7 - [Identification and Authentication Failures](https://owasp.org/Top10/A07_2021-Identification_and_Authentication_Failures/) (formerly Broken Authentication)\n- 2019 OWASP API Top 10 #2 - [Broken User Authentication](https://github.com/OWASP/API-Security/blob/master/2019/en/src/0xa2-broken-user-authentication.md)\n- 2019 OWASP API Top 10 #5 - [Broken Function Level Authorization](https://github.com/OWASP/API-Security/blob/master/2019/en/src/0xa5-broken-function-level-authorization.md)\n- 2022 CWE Top 25 #14 - [CWE-287: Improper Authentication](https://cwe.mitre.org/data/definitions/1387.html)\n- 2022 CWE Top 25 #16 - [CWE-862: Missing Authorization](https://cwe.mitre.org/data/definitions/1387.html)\n- 2022 CWE Top 25 #18 - [CWE-306: Missing Authentication for Critical Function](https://cwe.mitre.org/data/definitions/1387.html)\n- #21 most CVEs by CWE - [CWE-284: Access Control (Authorization) Issues](https://www.cvedetails.com/cwe-definitions.php)\n- #47 most CVEs by CWE - [CWE-639: Access Control Bypass Through User-Controlled Key](https://www.cvedetails.com/cwe-definitions.php)\n\nOf course, not all authn and authz bugs are due to route issues, but **`route-detect` seeks to automate detection of this vulnerability subclass.**\n\n![Routes demo](routes-demo.png?raw=true)\n\n<p align="center">\n    <i>Routes from <code><a href="https://github.com/koel/koel">koel<a></code> streaming server</i>\n</p>\n\nSupported web frameworks (`route-detect` IDs in parentheses):\n\n- Python: Django (`django`, `django-rest-framework`), Flask (`flask`), Sanic (`sanic`)\n- PHP: Laravel (`laravel`), Symfony (`symfony`), CakePHP (`cakephp`)\n- Ruby: Rails (`rails`), Grape (`grape`)\n- Java: JAX-RS (`jax-rs`), Spring (`spring`)\n- Go: Gorilla (`gorilla`), Gin (`gin`), Chi (`chi`)\n- JavaScript/TypeScript: Express (`express`), React (`react`), Angular (`angular`)\n\n# Installing\n\nUse `pip` to install `route-detect`:\n\n```\n$ python -m pip install --upgrade route-detect\n```\n\nYou can check that `route-detect` is installed correctly with the following command:\n\n```\n$ echo \'print(1 == 1)\' | semgrep --config $(routes which test-route-detect) -\nScanning 1 file.\n\nFindings:\n\n  /tmp/stdin\n     routes.rules.test-route-detect\n        Found \'1 == 1\', your route-detect installation is working correctly\n\n          1┆ print(1 == 1)\n\n\nRan 1 rule on 1 file: 1 finding.\n```\n\n# Using\n\n`route-detect` uses [`semgrep`](https://github.com/returntocorp/semgrep) to search for routes.\n\nUse the `which` command to point `semgrep` at the correct web application rules:\n\n```\n$ semgrep --config $(routes which django) path/to/django/code\n```\n\nUse the `viz` command to visualize route information in your browser:\n\n```\n$ semgrep --json --config $(routes which django) path/to/django/code > routes.json\n$ routes viz --browser routes.json\n```\n\n# Contributing\n\n`route-detect` uses [`poetry`](https://python-poetry.org/) for dependency and configuration management.\n\nBefore proceeding, install project dependencies with the following command:\n\n```\n$ poetry install --with dev\n```\n\n## Linting\n\nLint all project files with the following command:\n\n```\n$ poetry run pre-commit run --all-files\n```\n\n## Testing\n\nRun Python tests with the following command:\n\n```\n$ poetry run pytest --cov\n```\n\nRun Semgrep rule tests with the following command:\n\n```\n$ poetry run semgrep --test --config routes/rules/ tests/test_rules/\n```\n',
    'author': 'None',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mschwager/route-detect',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from . import commands  # noqa: F401
from . import rules  # noqa: F401
from . import templates  # noqa: F401


# Move to "importlib.metadata.version" when we only support Python 3.8+
__version__ = "0.5.0"
